#coding: utf-8

import bottle
import copy
import datetime
import json
import os
import requests
import time

from bottle import (post, route, run, request, redirect,
                    jinja2_template as template, response)
from cfg import get_base_config, update_config
from context import ContextStore
from raven import Client as RavenClient
from raven.contrib.bottle import Sentry
from redis import StrictRedis
from kairosapi.query import MetricsQuery
from log import logger, setup_logging
from os.path import dirname, abspath, join as join_path

bottle.TEMPLATE_PATH.insert(0, join_path(dirname(abspath(__file__)), 'views'))

setup_logging()
config = get_base_config()
if os.path.exists('config.yaml'):
    update_config(config, ['config.yaml'])

rdb = StrictRedis(**config['redis'])
ctx_store = ContextStore(rdb)


def get_kairosdb_metricnames(host, port=8080):
    resp = requests.get('http://{0}:{1}/api/v1/metricnames'.format(host, port), proxies={'http':''})
    resp.raise_for_status()
    return resp.json()['results']


@route('/list/<category>s')
def list_possible_categories(category):
    assert category in ['region', 'client', 'product', 'instance']
    response.content_type = 'application/json; charset=utf-8'
    return json.dumps(list(rdb.smembers('identities#{0}s'.format(category))))


@post('/incontext')
def query_in_context():
    def find_in_intervals(val, intervals, default=-1):
        for ind, (start, end) in enumerate(intervals):
            if start < val <= end:
                return ind
        else:
            return default

    bound_contexts = []
    ctx_spec = {}
    result_tags = []
    errors = []
    kairosdb_host = config['kairosdb']['host']
    kairosdb_port = config['kairosdb']['port']

    start = request.json.get('start', None)
    end = request.json.get('end', None)
    if not end:
        end = int(time.time())
    metric = request.json['metric']


    # Получим список существующих метрик от KairosDB и проверим,
    # что запрошенная вообще в них входит.
    # TODO: кэширование в Redis
    metrics = get_kairosdb_metricnames(kairosdb_host, kairosdb_port)
    if not metric in metrics:
        errors.append(('metric_not_found', 'Specified metric not found!'))

    if (end and end <= start):
        errors.append(('invalid_interval', 'End of interval should be greater than start!'))

    if not start:
        errors.append(('invalid_interval', 'Start of interval not specified!'))

    if errors:
        return {
            'errors': errors
        }

    # Если нам нужно дополнять получаемые данные какими-либо контекстами,
    # то получаем их из базы.
    if 'bound_contexts' in config and metric in config['bound_contexts']:
        ctx_spec = config['bound_contexts'][metric]
        bound_contexts = ctx_store.get(request.json['identity'],
                                       ctx_spec.keys(), start, end)

    tag_set = set()
    for ctx in bound_contexts:
        for part_name, field in ctx_spec.items():
            for v in ctx['info'][part_name]:
                tag_set.add(v)
    result_tags.extend(map(lambda v: {field: v}, tag_set))

    # Основу общих для всех запросов тэгов составляют опиисывающие
    # конкретный экземпляр продукта, по которому запрашиваются данные
    common_tags = copy.copy(request.json['identity'])
    kairosdb_url = 'http://{0}:{1}/api/v1/datapoints/query'.format(kairosdb_host,
                                                                   kairosdb_port)

    result_data = {
        'contexts': bound_contexts,
        'results': []
    }

    ctx_ts = map(lambda x: x['timestamp'], bound_contexts)
    ctx_ts.append(end)
    ts_bounds = zip(ctx_ts, ctx_ts[1:])

    # Составляем запрос к KairosDB
    metrics_q = MetricsQuery()
    metrics_q.set_start_from(start * 1000)
    metrics_q.set_end_to(end * 1000)

    # Пропишем агрегацию к запросу
    aggregators = []
    if 'aggregators' in request.json:
        for aggr_spec in request.json['aggregators']:
            fun, amount, period = aggr_spec.split('|')
            aggregators.append((fun, "{0} {1}".format(amount, period)))

    for specific_tags in result_tags:
        # Для каждого набора определяющих тегов необходимо составить отдельный запрос :(
        tags = copy.copy(common_tags)
        tags.update(specific_tags)
        # Если пользователь указал набор тегов, то его данные имеют приоритет
        tags.update(request.json.get('tags', {}))

        metrics_q.add_metric(metric, tags=tags, aggregators=aggregators)
        result_data['results'].append(dict(tags=tags, name=metric))
    else:
        metrics_q.add_metric(metric, aggregators=aggregators)

    # Отправляем запрос к KairosDB
    resp = requests.post(kairosdb_url,
                         proxies={'http': ''},
                         headers={'content-type': 'application/json'},
                         data=unicode(metrics_q))

    if resp.status_code == 200:
        json_data = json.loads(resp.content)
        raw_data = [d['results'][0] for d in json_data['queries']]
        # Дополним каждую метку ссылкой на временной контекст, в рамки которого она попадает.
        for i, portion in enumerate(raw_data[:-1]):
            result_data['results'][i]['values'] = map(
                                    lambda (ts, val): (ts, val) + (find_in_intervals(ts / 1000, ts_bounds, -1),),
                                    portion['values'])
    else:
        logger.error('Request to KairosDB failed: {0}@{1}\n{2}'.format(kairosdb_url,
                                                                       resp.status_code,
                                                                       unicode(metrics_q)))

    return result_data


@route('/debug')
def debug_view():
    keys = rdb.keys('context#*')
    contexts = []
    for set_key in keys:
        ctx = {
            'name': set_key,
            'saved': []
        }
        # Получаем общее число элементов в множестве контекста
        count = rdb.llen(set_key)
        items = rdb.lrange(set_key, 0, count+1)
        for text in items:
            content = eval(text)
            ts = datetime.datetime.fromtimestamp(content['timestamp'])
            ctx['saved'].append({
                'date': ts.strftime('%H:%M:%S %d.%m.%Y'),
                'ts': int(content['timestamp']),
                'content': json.dumps(content, indent=4)
                })
        contexts.append(ctx)
    return template('debug_list.tpl', contexts=contexts)


@route('/')
def index():
    return redirect('/debug')


@post('/register')
def submit_info():
    ts = request.json['timestamp']
    info = request.json['info']
    # Для оптимизации последующей обработки, каждая составляющая общего
    # контекста сохраняется в отдельном множестве
    for key, val in info.iteritems():
        ctx_store.put(request.json['identity'], key, val, ts)
    # Сохраняем части названий для отдачи полям выбора в web-интерфейсе.
    for k, v in request.json['identity'].iteritems():
        rdb.sadd('identities#{0}s'.format(k), v)
    return {'result': 'OK'}


app = bottle.default_app()

if 'raven' in config:
    client = RavenClient(**config['raven'])
    app.catchall = False
    app = Sentry(app, client)

if __name__ == '__main__':
    run(app=app, **config['app'])
